import logging
from pathlib import Path

import pcbnew

from .common import (
    calculate_translation,
    flip_dict,
    get_group_footprints,
    get_layout_map,
    sync_footprints,
    sync_track,
    sync_zone,
    update_zone_net,
)

log = logging.getLogger(__name__)


class PullGroup(pcbnew.ActionPlugin):
    def defaults(self):
        self.name = "Pull Group"
        self.category = "Pull Group Layout Atopile"
        self.description = "Layout components on PCB in same spatial relationships as components on schematic"
        self.show_toolbar_button = True
        self.icon_file_name = str(Path(__file__).parent / "download.png")
        self.dark_icon_file_name = self.icon_file_name

    def Run(self):
        target_board: pcbnew.BOARD = pcbnew.GetBoard()
        board_path = target_board.GetFileName()
        known_layouts = get_layout_map(board_path)

        # Pull Selected Groups
        for g in target_board.Groups():
            assert isinstance(g, pcbnew.PCB_GROUP)

            # Skip over unselected groups
            if not g.IsSelected():
                continue

            g_name = g.GetName()
            if g_name not in known_layouts:
                continue

            layout_path = Path(known_layouts[g_name]["layout_path"])
            # Check what layouts exist
            if not layout_path.exists():
                raise FileNotFoundError(
                    f"Cannot load group. Layout file {layout_path} does not exist"
                )

            # Remove everything but the footprints from the target group
            for item in g.GetItems():
                if not isinstance(item, pcbnew.FOOTPRINT):
                    target_board.Remove(item)

            # Load the layout and sync
            source_board: pcbnew.BOARD = pcbnew.LoadBoard(str(layout_path))

            # Calculate offset before moving footprints
            offset = calculate_translation(source_fps=source_board.GetFootprints(), target_fps=get_group_footprints(g))

            sync_footprints(
                source_board, target_board, flip_dict(known_layouts[g_name]["uuid_map"])
            )

            for track in source_board.GetTracks():
                item = sync_track(source_board, track, target_board)
                g.AddItem(item)

            for zone in source_board.Zones():
                new_zone = sync_zone(zone,target_board)
                update_zone_net(zone, source_board, new_zone, target_board, flip_dict(known_layouts[g_name]["uuid_map"]))
                g.AddItem(new_zone)

            # Shift entire target group by offset as last operation
            g.Move(offset)

        pcbnew.Refresh()


PullGroup().register()



import os
import shutil
import stat
from pathlib import Path


def robustly_rm_dir(path: Path) -> None:
    """Remove a directory and all its contents."""

    def remove_readonly(func, path, excinfo):
        os.chmod(path, stat.S_IWRITE)
        func(path)

    shutil.rmtree(path, onerror=remove_readonly)



"""
Work with expressions and ranged values.
"""

import collections.abc
from typing import Callable, Mapping, Optional, Type, Union

import pint
from attrs import define, frozen
from pint.facets.plain import PlainUnit

_UNITLESS = pint.Unit("")


def _custom_float_format(value, max_decimals: int):
    """
    Format the float to have up to max_decimals places, but fewer if there's no more precision.
    """
    # Format with fixed-point notation, up to max_decimals
    formatted = f"{value:.{max_decimals}f}"
    # Remove trailing zeros and the decimal point if not needed
    return formatted.rstrip("0").rstrip(".")


def _best_units(qty_a: pint.Quantity, qty_b: pint.Quantity) -> PlainUnit:
    """Return the best unit for the two quantities."""
    if str(qty_a.to(qty_b.units).magnitude) < str(qty_b.to(qty_a.units).magnitude):
        return qty_a.units
    return qty_b.units


_favourable_units = [
    pint.Unit("V"),
    pint.Unit("ohm"),
    pint.Unit("A"),
    pint.Unit("W"),
    pint.Unit("Hz"),
    pint.Unit("F"),
    pint.Unit("H"),
]


def _favourable_unit(unit: PlainUnit) -> PlainUnit:
    """Return the most favourable unit for the given unit."""
    for fav_unit in _favourable_units:
        if unit.is_compatible_with(fav_unit):
            return fav_unit
    return unit


class RangedValue:
    """
    Let's get physical!

    Ranged values are designed to represent a range of physical values, such as
    a voltage or current, including tolerances.
    """

    def __init__(
        self,
        val_a: Union[float, int, pint.Quantity],
        val_b: Optional[Union[float, int, pint.Quantity]] = None,
        unit: Optional[str | PlainUnit | pint.Unit] = None,
        pretty_unit: Optional[str] = None,
    ):
        # This is a bit of a hack, but simplifies upstream code marginally
        if val_b is None:
            val_b = val_a

        # If we're given a unit, use it. Otherwise, try to infer the unit from the inputs.
        if unit:
            self.unit = pint.Unit(unit)
        elif isinstance(val_a, pint.Quantity) and isinstance(val_b, pint.Quantity):
            self.unit = _best_units(val_a, val_b)
        elif isinstance(val_a, pint.Quantity):
            self.unit = val_a.units
        elif isinstance(val_a, pint.Quantity):
            self.unit = val_b.units
        else:
            self.unit = _UNITLESS

        # If the inputs are pint Quantities, convert them to the same unit
        if isinstance(val_a, pint.Quantity):
            val_a_mag = val_a.to(self.unit).magnitude
        else:
            val_a_mag = val_a

        if isinstance(val_b, pint.Quantity):
            val_b_mag = val_b.to(self.unit).magnitude
        else:
            val_b_mag = val_b

        assert isinstance(val_a_mag, (float, int))
        assert isinstance(val_b_mag, (float, int))

        # Make the noise
        self.pretty_unit = pretty_unit
        self.min_val = min(val_a_mag, val_b_mag)
        self.max_val = max(val_a_mag, val_b_mag)

    @property
    def best_usr_unit(self) -> str:
        """Return a pretty string representation of the unit."""
        if self.pretty_unit:
            return self.pretty_unit
        return str(self.unit)

    def to(self, unit: str | PlainUnit | pint.Unit) -> "RangedValue":
        """Return a new RangedValue in the given unit."""
        return RangedValue(
            self.min_qty,
            self.max_qty,
            unit
        )

    def to_compact(self) -> "RangedValue":
        """Return a new RangedValue in the most compact unit."""
        return RangedValue(
            # FIXME: still shit
            self.min_qty.to_compact(),
            self.max_qty.to_compact(),
        )

    def pretty_str(
        self,
        max_decimals: Optional[int] = 2,
        unit: Optional[pint.Unit] = None
    ) -> str:
        """Return a pretty string representation of the RangedValue."""
        if unit is not None:
            val = self.to(unit)
        else:
            val = self.to(_favourable_unit(self.unit)).to_compact()

        if max_decimals is None:
            nom = str(val.nominal)
            if val.tolerance != 0:
                nom += f' +/- {str(val.tolerance)}'
                if val.tolerance_pct is not None:
                    nom += f' ({str(val.tolerance_pct)}%)'
        else:
            nom = _custom_float_format(val.nominal, max_decimals)
            if val.tolerance != 0:
                nom += f' +/- {_custom_float_format(val.tolerance, max_decimals)}'
                if val.tolerance_pct is not None:
                    nom += f' ({_custom_float_format(val.tolerance_pct, max_decimals)}%)'

        return f"{nom} {val.best_usr_unit}"

    def __str__(self) -> str:
        return self.pretty_str()

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}({self.min_val}, {self.max_val}, '{self.unit}')"
        )

    @property
    def nominal(self) -> float:
        """Return the nominal value of the RangedValue."""
        return (self.min_val + self.max_val) / 2

    @property
    def tolerance(self) -> float:
        """Return the tolerance of the RangedValue."""
        return (self.max_val - self.min_val) / 2

    @property
    def tolerance_pct(self) -> Optional[float]:
        """Return the tolerance as a percentage of the nominal value."""
        if self.nominal == 0:
            return None
        return self.tolerance / self.nominal * 100

    def to_dict(self) -> dict:
        """Convert the Physical instance to a dictionary."""
        return {
            "unit": str(self.unit),
            "min_val": self.min_val,
            "max_val": self.max_val,
            # TODO: remove these - we shouldn't be duplicating this kind of information
            "nominal": self.nominal,
            "tolerance": self.tolerance,
            "tolerance_pct": self.tolerance_pct,
        }

    @property
    def min_qty(self) -> pint.Quantity:
        """Return the minimum of self as a pint Quantity."""
        return self.unit * self.min_val

    @property
    def max_qty(self) -> pint.Quantity:
        """Return the maximum of self as a pint Quantity."""
        return self.unit * self.max_val

    @classmethod
    def _ensure(cls, thing) -> "RangedValue":
        if isinstance(thing, RangedValue):
            return thing
        return cls(thing, thing, _UNITLESS)

    def __mul__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        other = self._ensure(other)

        new_values = [
            self.min_qty * other.min_qty,
            self.min_qty * other.max_qty,
            self.max_qty * other.min_qty,
            self.max_qty * other.max_qty,
        ]

        return self.__class__(
            min(new_values),
            max(new_values),
        )

    def __rmul__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        return self.__mul__(other)

    def __pow__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        if isinstance(other, RangedValue):
            if not (other.unit.dimensionless and other.min_val == other.max_val):
                raise ValueError("Exponent must be a constant valueless quantity")
            other = other.min_val

        return self.__class__(self.min_qty**other, self.max_qty**other)

    @classmethod
    def _do_truediv(
        cls,
        numerator: Union["RangedValue", float, int],
        denominator: Union["RangedValue", float, int],
    ) -> "RangedValue":
        numerator = cls._ensure(numerator)
        denominator = cls._ensure(denominator)

        new_values = [
            numerator.min_qty / denominator.min_qty,
            numerator.min_qty / denominator.max_qty,
            numerator.max_qty / denominator.min_qty,
            numerator.max_qty / denominator.max_qty,
        ]

        return cls(
            min(new_values),
            max(new_values),
        )

    def __truediv__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        return self._do_truediv(self, other)

    def __rtruediv__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        return self._do_truediv(other, self)

    def __add__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        other = self._ensure(other)

        return self.__class__(
            self.min_qty + other.min_qty,
            self.max_qty + other.max_qty,
        )

    def __radd__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        return self.__add__(other)

    def __sub__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        other = self._ensure(other)

        return self.__class__(
            self.min_qty - other.max_qty,
            self.max_qty - other.min_qty,
        )

    def __rsub__(self, other: Union["RangedValue", float, int]) -> "RangedValue":
        return self.__sub__(other)

    def __neg__(self) -> "RangedValue":
        return self.__class__(
            -self.max_qty, -self.min_qty, self.unit, self.pretty_unit
        )

    def within(self, other: Union["RangedValue", float, int]) -> bool:
        """Check that this RangedValue completely falls within another."""
        if not isinstance(other, RangedValue):
            if not self.unit.dimensionless:
                raise ValueError(
                    "Can only compare RangedValue to a dimensionless quantity"
                )
            return self.min_val == self.max_val == other
        return self.min_qty >= other.min_qty and other.max_qty >= self.max_qty

    # NOTE: we use the < and > operators interchangeably with the <= and >= operators
    def __lt__(self, other: Union["RangedValue", float, int]) -> bool:
        other = self._ensure(other)
        return self.max_qty <= other.min_qty

    def __gt__(self, other: Union["RangedValue", float, int]) -> bool:
        other = self._ensure(other)
        return self.min_qty >= other.max_qty

    def __eq__(self, other: object) -> bool:
        # NOTE: realistically this is only useful for testing
        if isinstance(other, RangedValue):
            return self.min_qty == other.min_qty and self.max_qty == other.max_qty
        if self.min_val == self.max_val == other and self.unit.dimensionless:
            return True
        return False

    def __req__(self, other: Union["RangedValue", float, int]) -> bool:
        return self.__eq__(other)


NumericishTypes = Union["Expression", RangedValue, float, int, "Symbol"]


@frozen
class Symbol:
    """Represent a symbol."""

    key: collections.abc.Hashable

    def __call__(self, context: Mapping) -> RangedValue:
        """Return the value of the symbol."""
        thing = context[self.key]
        if callable(thing):
            return thing(context)
        return thing


# TODO: figure out how to pretty print these with the symbols etc...
@define
class Expression:
    """Represent an expression."""

    symbols: set[Symbol]
    lambda_: Callable[[Mapping[str, NumericishTypes]], RangedValue]

    @classmethod
    def from_expr(cls, expr: "Expression") -> "Expression":
        """Create an expression from another expression."""
        return cls(symbols=expr.symbols, lambda_=expr.lambda_)

    @classmethod
    def from_numericish(cls, thing: NumericishTypes) -> "Expression":
        """Create an expression from a numericish thing."""
        if isinstance(thing, Expression):
            return cls.from_expr(thing)
        if isinstance(thing, RangedValue):
            return cls(symbols=set(), lambda_=lambda context: thing)
        if isinstance(thing, Symbol):
            return cls(symbols={thing}, lambda_=thing)
        return cls(symbols=set(), lambda_=lambda context: RangedValue(thing, thing))

    def __call__(self, context: Mapping[str, NumericishTypes]) -> RangedValue:
        return self.lambda_(context)


def _get_symbols(thing: NumericishTypes) -> set[Symbol]:
    if isinstance(thing, Expression):
        return thing.symbols
    elif isinstance(thing, Symbol):
        return {thing}
    else:
        return set()


def defer_operation_factory(
    lhs: NumericishTypes,
    operator: Callable,
    rhs: NumericishTypes,
    deffering_type: Type = Expression,
) -> NumericishTypes:
    """Create a deferred operation, using deffering_type as the base for teh callable."""
    if not callable(lhs) and not callable(rhs):
        # in this case we can just do the operation now, skip ahead and merry christmas
        return operator(lhs, rhs)

    # if we're here, we need to create an expression
    symbols = _get_symbols(lhs) | _get_symbols(rhs)
    if callable(lhs) and callable(rhs):

        def lambda_(context):
            return operator(lhs(context), rhs(context))

    elif callable(lhs):

        def lambda_(context):
            return operator(lhs(context), rhs)

    else:

        def lambda_(context):
            return operator(lhs, rhs(context))

    return deffering_type(symbols=symbols, lambda_=lambda_)



# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
"""Implementation of tool support over LSP."""
from __future__ import annotations

import copy
import json
import os
import pathlib
import sys
import traceback
from pathlib import Path
from typing import Any, Optional, Sequence

import atopile.address
import atopile.config
import atopile.datatypes
import atopile.errors
import atopile.front_end
import atopile.parse
import atopile.parse_utils

# **********************************************************
# Utils for interacting with the atopile front-end
# **********************************************************

_line_to_def_block: dict[Path, dict[int, atopile.address.AddrStr]] = {}


def _reset_caches(file: Path):
    """Remove a file from the cache."""
    if file in atopile.parse.parser.cache:
        del atopile.parse.parser.cache[file]

    if file in _line_to_def_block:
        del _line_to_def_block[file]

    # TODO: only clear these caches of what's been invalidated
    file_str = str(file)
    to_clear = [addr for addr in atopile.front_end.scoop._output_cache if addr.startswith(file_str)]
    for addr in to_clear:
        del atopile.front_end.scoop._output_cache[addr]

    to_clear = [addr for addr in atopile.front_end.dizzy._output_cache if addr.startswith(file_str)]
    for addr in to_clear:
        del atopile.front_end.dizzy._output_cache[addr]

    atopile.front_end.lofty._output_cache.clear()


def _index_class_defs_by_line(file: Path):
    """Index class definitions in a given file by the line number"""
    _line_to_def_block[file] = {}
    addrs = atopile.front_end.scoop.ingest_file(file)

    for addr in addrs:
        if addr == str(file):
            continue
        try:
            atopile.front_end.lofty.get_instance(addr)
        except atopile.errors.AtoError as ex:
            log_warning(str(ex))
            continue

        # FIXME: we shouldn't be entangling this
        # code w/ the front-end so much
        try:
            cls_def = atopile.front_end.scoop.get_obj_def(addr)
            _, start_line, _, stop_line, _ = atopile.parse_utils.get_src_info_from_ctx(
                cls_def.src_ctx.block()
            )
        except AttributeError:
            continue

        for i in range(start_line + 1, stop_line + 1):
            _line_to_def_block[file][i] = cls_def.address


def _get_def_addr_from_line(file: Path, line: int) -> Optional[atopile.address.AddrStr]:
    """Get the class definition from a line number"""
    if file not in _line_to_def_block or not _line_to_def_block[file]:
        _index_class_defs_by_line(file)
    return _line_to_def_block[file].get(line, None)


# **********************************************************
# Update sys.path before importing any bundled libraries.
# **********************************************************
def update_sys_path(path_to_add: str, strategy: str) -> None:
    """Add given path to `sys.path`."""
    if path_to_add not in sys.path and os.path.isdir(path_to_add):
        if strategy == "useBundled":
            sys.path.insert(0, path_to_add)
        elif strategy == "fromEnvironment":
            sys.path.append(path_to_add)


# Ensure that we can import LSP libraries, and other bundled libraries.
update_sys_path(
    os.fspath(pathlib.Path(__file__).parent.parent / "libs"),
    os.getenv("LS_IMPORT_STRATEGY", "useBundled"),
)

# **********************************************************
# Imports needed for the language server goes below this.
# **********************************************************
# pylint: disable=wrong-import-position,import-error
import lsp_jsonrpc as jsonrpc  # noqa: E402
import lsp_utils as utils  # noqa: E402
import lsprotocol.types as lsp  # noqa: E402
from pygls import server, uris, workspace  # noqa: E402

WORKSPACE_SETTINGS = {}
GLOBAL_SETTINGS = {}
RUNNER = pathlib.Path(__file__).parent / "lsp_runner.py"

MAX_WORKERS = 5
# TODO: Update the language server name and version.
LSP_SERVER = server.LanguageServer(
    name="atopile", version="<server version>", max_workers=MAX_WORKERS
)


# **********************************************************
# Tool specific code goes below this.
# **********************************************************

# Reference:
#  LS Protocol:
#  https://microsoft.github.io/language-server-protocol/specifications/specification-3-16/
#
#  Sample implementations:
#  Pylint: https://github.com/microsoft/vscode-pylint/blob/main/bundled/tool
#  Black: https://github.com/microsoft/vscode-black-formatter/blob/main/bundled/tool
#  isort: https://github.com/microsoft/vscode-isort/blob/main/bundled/tool

# e.g, TOOL_MODULE = "pylint"
TOOL_MODULE = "atopile"

# e.g, TOOL_DISPLAY = "Pylint"
TOOL_DISPLAY = "atopile"

# all scenarios.
TOOL_ARGS = []  # default arguments always passed to your tool.


# TODO: If your tool is a linter then update this section.
# Delete "Linting features" section if your tool is NOT a linter.
# **********************************************************
# Linting features start here
# **********************************************************

#  See `pylint` implementation for a full featured linter extension:
#  Pylint: https://github.com/microsoft/vscode-pylint/blob/main/bundled/tool


@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_COMPLETION, lsp.CompletionOptions(trigger_characters=["."]),)
def completions(params: Optional[lsp.CompletionParams] = None) -> lsp.CompletionList:
    """Handler for completion requests."""
    if not params.text_document.uri.startswith("file://"):
        return lsp.CompletionList(is_incomplete=False, items=[])

    document = LSP_SERVER.workspace.get_text_document(params.text_document.uri)

    file = Path(document.path)
    try:
        atopile.config.get_project_context()
    except ValueError:
        atopile.config.set_project_context(atopile.config.ProjectContext.from_path(file))

    word = utils.cursor_word(document, params.position)
    class_addr = _get_def_addr_from_line(file, params.position.line) or str(file)
    items: list[lsp.CompletionItem] = []

    instance_addr = atopile.address.add_instances(class_addr, word.split(".")[:-1])
    try:
        instance = atopile.front_end.lofty.get_instance(instance_addr)
    except (KeyError, atopile.errors.AtoError):
        pass
    else:
        for child, assignment in instance.assignments.items():
            items.append(lsp.CompletionItem(label=child, kind=lsp.CompletionItemKind.Property, detail=assignment[0].given_type))

        for child in instance.children:
            items.append(lsp.CompletionItem(label=child, kind=lsp.CompletionItemKind.Method))

    # Class Defs
    # FIXME: this is a hack to check whether something could not be class-def
    if "." not in word:
        try:
            class_ctx = atopile.front_end.scoop.get_obj_def(class_addr)
        except (KeyError, atopile.errors.AtoError):
            pass
        else:
            closure_contexts = [class_ctx]
            if class_ctx.closure:
                closure_contexts.extend(class_ctx.closure)

            for closure_ctx in closure_contexts:

                for cls_ref in closure_ctx.local_defs:
                    items.append(lsp.CompletionItem(label=cls_ref[0], kind=lsp.CompletionItemKind.Class))

                for imp_ref in closure_ctx.imports:
                    items.append(lsp.CompletionItem(label=imp_ref[0], kind=lsp.CompletionItemKind.Class))

    return lsp.CompletionList(is_incomplete=False, items=items,)


@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DEFINITION)
def goto_definition(params: Optional[lsp.DefinitionParams] = None) -> Optional[lsp.Location]:
    """Handler for goto definition."""
    if not params.text_document.uri.startswith("file://"):
        return lsp.CompletionList(is_incomplete=False, items=[])

    document = LSP_SERVER.workspace.get_text_document(params.text_document.uri)

    file = Path(document.path)
    try:
        atopile.config.get_project_context()
    except ValueError:
        atopile.config.set_project_context(atopile.config.ProjectContext.from_path(file))

    class_addr = _get_def_addr_from_line(file, params.position.line)
    if not class_addr:
        class_addr = str(file)

    word, range_ = utils.cursor_word_and_range(document, params.position)
    try:
        word = word[:word.index(".", params.position.character - range_.start.character)]
    except ValueError:
        pass

    # See if it's an instance
    instance_addr = atopile.address.add_instances(class_addr, word.split("."))
    src_ctx = None
    try:
        src_ctx = atopile.front_end.lofty.get_instance(instance_addr).src_ctx
    except (KeyError, atopile.errors.AtoError, AttributeError):
        # See if it's a Class instead
        pass

    # See if it's a class
    try:
        src_ctx = atopile.front_end.scoop.get_obj_def(
            atopile.front_end.lookup_class_in_closure(
                atopile.front_end.scoop.get_obj_def(class_addr),
                atopile.datatypes.Ref(word.split("."))
            )
        ).src_ctx
    except (KeyError, atopile.errors.AtoError, AttributeError):
        pass

    try:
        file_path, start_line, start_col, stop_line, stop_col = atopile.parse_utils.get_src_info_from_ctx(src_ctx)
    except AttributeError:
        return
    else:
        return lsp.Location(
            "file://" + str(file_path),
            lsp.Range(
                lsp.Position(start_line - 1, start_col),
                lsp.Position(stop_line - 1, stop_col)
            )
        )


# TODO: we don't currently have good enough cache management to support dynamic parsing
# @LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DID_CHANGE)
# def did_change(params: lsp.DidChangeTextDocumentParams) -> None:
#     """LSP handler for textDocument/didOpen request."""
#     document = LSP_SERVER.workspace.get_text_document(params.text_document.uri)

#     # naive method to real-time lint on every keypress
#     _linting_helper(document)


@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DID_OPEN)
def did_open(params: lsp.DidOpenTextDocumentParams) -> None:
    """LSP handler for textDocument/didOpen request."""
    if not params.text_document.uri.startswith("file://"):
        return lsp.CompletionList(is_incomplete=False, items=[])

    document = LSP_SERVER.workspace.get_text_document(params.text_document.uri)

    file = Path(document.path)
    try:
        atopile.config.get_project_context()
    except ValueError:
        atopile.config.set_project_context(atopile.config.ProjectContext.from_path(file))

    _index_class_defs_by_line(file)


# TODO: remove me if we don't have a purpose for it
# @LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DID_CLOSE)
# def did_close(params: lsp.DidCloseTextDocumentParams) -> None:
#     """LSP handler for textDocument/didClose request."""
#     #document = LSP_SERVER.workspace.get_text_document(params.text_document.uri)
#     # Publishing empty diagnostics to clear the entries for this file.
#     #LSP_SERVER.publish_diagnostics(document.uri, [])
#     log_to_output("did close")


@LSP_SERVER.feature(lsp.TEXT_DOCUMENT_DID_SAVE)
def did_save(params: lsp.DidSaveTextDocumentParams) -> None:
    """LSP handler for textDocument/didSave request."""
    if not params.text_document.uri.startswith("file://"):
        return lsp.CompletionList(is_incomplete=False, items=[])

    document = LSP_SERVER.workspace.get_text_document(params.text_document.uri)

    file = Path(document.path)

    _reset_caches(file)
    _index_class_defs_by_line(file)


# TODO: if you want to handle setting specific severity for your linter
# in a user configurable way, then look at look at how it is implemented
# for `pylint` extension from our team.
# Pylint: https://github.com/microsoft/vscode-pylint
# Follow the flow of severity from the settings in package.json to the server.
def _get_severity(*_codes: list[str]) -> lsp.DiagnosticSeverity:
    # TODO: All reported issues from linter are treated as warning.
    # change it as appropriate for your linter.
    return lsp.DiagnosticSeverity.Error



# **********************************************************
# Required Language Server Initialization and Exit handlers.
# **********************************************************
@LSP_SERVER.feature(lsp.INITIALIZE)
def initialize(params: lsp.InitializeParams) -> None:
    """LSP handler for initialize request."""
    log_to_output(f"CWD Server: {os.getcwd()}")

    paths = "\r\n   ".join(sys.path)
    log_to_output(f"sys.path used to run Server:\r\n   {paths}")

    GLOBAL_SETTINGS.update(**params.initialization_options.get("globalSettings", {}))

    settings = params.initialization_options["settings"]
    _update_workspace_settings(settings)
    log_to_output(
        f"Settings used to run Server:\r\n{json.dumps(settings, indent=4, ensure_ascii=False)}\r\n"
    )
    log_to_output(
        f"Global settings:\r\n{json.dumps(GLOBAL_SETTINGS, indent=4, ensure_ascii=False)}\r\n"
    )


@LSP_SERVER.feature(lsp.EXIT)
def on_exit(_params: Optional[Any] = None) -> None:
    """Handle clean up on exit."""
    jsonrpc.shutdown_json_rpc()


@LSP_SERVER.feature(lsp.SHUTDOWN)
def on_shutdown(_params: Optional[Any] = None) -> None:
    """Handle clean up on shutdown."""
    jsonrpc.shutdown_json_rpc()


def _get_global_defaults():
    return {
        "path": GLOBAL_SETTINGS.get("path", []),
        "interpreter": GLOBAL_SETTINGS.get("interpreter", [sys.executable]),
        "args": GLOBAL_SETTINGS.get("args", []),
        "importStrategy": GLOBAL_SETTINGS.get("importStrategy", "useBundled"),
        "showNotifications": GLOBAL_SETTINGS.get("showNotifications", "off"),
    }


def _update_workspace_settings(settings):
    if not settings:
        key = os.getcwd()
        WORKSPACE_SETTINGS[key] = {
            "cwd": key,
            "workspaceFS": key,
            "workspace": uris.from_fs_path(key),
            **_get_global_defaults(),
        }
        return

    for setting in settings:
        key = uris.to_fs_path(setting["workspace"])
        WORKSPACE_SETTINGS[key] = {
            "cwd": key,
            **setting,
            "workspaceFS": key,
        }


def _get_settings_by_path(file_path: pathlib.Path):
    workspaces = {s["workspaceFS"] for s in WORKSPACE_SETTINGS.values()}

    while file_path != file_path.parent:
        str_file_path = str(file_path)
        if str_file_path in workspaces:
            return WORKSPACE_SETTINGS[str_file_path]
        file_path = file_path.parent

    setting_values = list(WORKSPACE_SETTINGS.values())
    return setting_values[0]


def _get_document_key(document: workspace.Document):
    if WORKSPACE_SETTINGS:
        document_workspace = pathlib.Path(document.path)
        workspaces = {s["workspaceFS"] for s in WORKSPACE_SETTINGS.values()}

        # Find workspace settings for the given file.
        while document_workspace != document_workspace.parent:
            if str(document_workspace) in workspaces:
                return str(document_workspace)
            document_workspace = document_workspace.parent

    return None


def _get_settings_by_document(document: workspace.Document | None):
    if document is None or document.path is None:
        return list(WORKSPACE_SETTINGS.values())[0]

    key = _get_document_key(document)
    if key is None:
        # This is either a non-workspace file or there is no workspace.
        key = os.fspath(pathlib.Path(document.path).parent)
        return {
            "cwd": key,
            "workspaceFS": key,
            "workspace": uris.from_fs_path(key),
            **_get_global_defaults(),
        }

    return WORKSPACE_SETTINGS[str(key)]


# *****************************************************
# Internal execution APIs.
# *****************************************************
def _run_tool_on_document(
    document: workspace.Document,
    use_stdin: bool = False,
    extra_args: Optional[Sequence[str]] = None,
) -> utils.RunResult | None:
    """Runs tool on the given document.

    if use_stdin is true then contents of the document is passed to the
    tool via stdin.
    """
    if extra_args is None:
        extra_args = []
    if str(document.uri).startswith("vscode-notebook-cell"):
        # TODO: Decide on if you want to skip notebook cells.
        # Skip notebook cells
        return None

    if utils.is_stdlib_file(document.path):
        # TODO: Decide on if you want to skip standard library files.
        # Skip standard library python files.
        return None

    # deep copy here to prevent accidentally updating global settings.
    settings = copy.deepcopy(_get_settings_by_document(document))

    code_workspace = settings["workspaceFS"]
    cwd = settings["cwd"]

    use_path = False
    use_rpc = False
    if settings["path"]:
        # 'path' setting takes priority over everything.
        use_path = True
        argv = settings["path"]
    elif settings["interpreter"] and not utils.is_current_interpreter(
        settings["interpreter"][0]
    ):
        # If there is a different interpreter set use JSON-RPC to the subprocess
        # running under that interpreter.
        argv = [TOOL_MODULE]
        use_rpc = True
    else:
        # if the interpreter is same as the interpreter running this
        # process then run as module.
        argv = [TOOL_MODULE]

    argv += TOOL_ARGS + settings["args"] + extra_args

    if use_stdin:
        # TODO: update these to pass the appropriate arguments to provide document contents
        # to tool via stdin.
        # For example, for pylint args for stdin looks like this:
        #     pylint --from-stdin <path>
        # Here `--from-stdin` path is used by pylint to make decisions on the file contents
        # that are being processed. Like, applying exclusion rules.
        # It should look like this when you pass it:
        #     argv += ["--from-stdin", document.path]
        # Read up on how your tool handles contents via stdin. If stdin is not supported use
        # set use_stdin to False, or provide path, what ever is appropriate for your tool.
        argv += []
    else:
        argv += [document.path]

    if use_path:
        # This mode is used when running executables.
        log_to_output(" ".join(argv))
        log_to_output(f"CWD Server: {cwd}")
        result = utils.run_path(
            argv=argv,
            use_stdin=use_stdin,
            cwd=cwd,
            source=document.source.replace("\r\n", "\n"),
        )
        if result.stderr:
            log_to_output(result.stderr)
    elif use_rpc:
        # This mode is used if the interpreter running this server is different from
        # the interpreter used for running this server.
        log_to_output(" ".join(settings["interpreter"] + ["-m"] + argv))
        log_to_output(f"CWD Linter: {cwd}")

        result = jsonrpc.run_over_json_rpc(
            workspace=code_workspace,
            interpreter=settings["interpreter"],
            module=TOOL_MODULE,
            argv=argv,
            use_stdin=use_stdin,
            cwd=cwd,
            source=document.source,
        )
        if result.exception:
            log_error(result.exception)
            result = utils.RunResult(result.stdout, result.stderr)
        elif result.stderr:
            log_to_output(result.stderr)
    else:
        # In this mode the tool is run as a module in the same process as the language server.
        log_to_output(" ".join([sys.executable, "-m"] + argv))
        log_to_output(f"CWD Linter: {cwd}")
        # This is needed to preserve sys.path, in cases where the tool modifies
        # sys.path and that might not work for this scenario next time around.

        # with utils.substitute_attr(sys, "path", sys.path[:]):
        #     try:
        #         # TODO: `utils.run_module` is equivalent to running `python -m atopile`.
        #         # If your tool supports a programmatic API then replace the function below
        #         # with code for your tool. You can also use `utils.run_api` helper, which
        #         # handles changing working directories, managing io streams, etc.
        #         # Also update `_run_tool` function and `utils.run_module` in `lsp_runner.py`.
        #         result = utils.run_module(
        #             module=TOOL_MODULE,
        #             argv=argv,
        #             use_stdin=use_stdin,
        #             cwd=cwd,
        #             source=document.source,
        #         )
        #     except Exception:
        #         log_error(traceback.format_exc(chain=True))
        #         raise

        if result.stderr:
            log_to_output(result.stderr)

    log_to_output(f"{document.uri} :\r\n{result.stdout}")
    return result


def _run_tool(extra_args: Sequence[str]) -> utils.RunResult:
    """Runs tool."""
    # deep copy here to prevent accidentally updating global settings.
    settings = copy.deepcopy(_get_settings_by_document(None))

    code_workspace = settings["workspaceFS"]
    cwd = settings["workspaceFS"]

    use_path = False
    use_rpc = False
    if len(settings["path"]) > 0:
        # 'path' setting takes priority over everything.
        use_path = True
        argv = settings["path"]
    elif len(settings["interpreter"]) > 0 and not utils.is_current_interpreter(
        settings["interpreter"][0]
    ):
        # If there is a different interpreter set use JSON-RPC to the subprocess
        # running under that interpreter.
        argv = [TOOL_MODULE]
        use_rpc = True
    else:
        # if the interpreter is same as the interpreter running this
        # process then run as module.
        argv = [TOOL_MODULE]

    argv += extra_args

    if use_path:
        # This mode is used when running executables.
        log_to_output(" ".join(argv))
        log_to_output(f"CWD Server: {cwd}")
        result = utils.run_path(argv=argv, use_stdin=True, cwd=cwd)
        if result.stderr:
            log_to_output(result.stderr)
    elif use_rpc:
        # This mode is used if the interpreter running this server is different from
        # the interpreter used for running this server.
        log_to_output(" ".join(settings["interpreter"] + ["-m"] + argv))
        log_to_output(f"CWD Linter: {cwd}")
        result = jsonrpc.run_over_json_rpc(
            workspace=code_workspace,
            interpreter=settings["interpreter"],
            module=TOOL_MODULE,
            argv=argv,
            use_stdin=True,
            cwd=cwd,
        )
        if result.exception:
            log_error(result.exception)
            result = utils.RunResult(result.stdout, result.stderr)
        elif result.stderr:
            log_to_output(result.stderr)
    else:
        # In this mode the tool is run as a module in the same process as the language server.
        log_to_output(" ".join([sys.executable, "-m"] + argv))
        log_to_output(f"CWD Linter: {cwd}")
        # This is needed to preserve sys.path, in cases where the tool modifies
        # sys.path and that might not work for this scenario next time around.
        with utils.substitute_attr(sys, "path", sys.path[:]):
            try:
                # TODO: `utils.run_module` is equivalent to running `python -m atopile`.
                # If your tool supports a programmatic API then replace the function below
                # with code for your tool. You can also use `utils.run_api` helper, which
                # handles changing working directories, managing io streams, etc.
                # Also update `_run_tool_on_document` function and `utils.run_module` in `lsp_runner.py`.
                result = utils.run_module(
                    module=TOOL_MODULE, argv=argv, use_stdin=True, cwd=cwd
                )
            except Exception:
                log_error(traceback.format_exc(chain=True))
                raise
        if result.stderr:
            log_to_output(result.stderr)

    log_to_output(f"\r\n{result.stdout}\r\n")
    return result


# *****************************************************
# Logging and notification.
# *****************************************************
def log_to_output(
    message: str, msg_type: lsp.MessageType = lsp.MessageType.Log
) -> None:
    LSP_SERVER.show_message_log(message, msg_type)


def log_error(message: str) -> None:
    LSP_SERVER.show_message_log(message, lsp.MessageType.Error)
    if os.getenv("LS_SHOW_NOTIFICATION", "off") in ["onError", "onWarning", "always"]:
        LSP_SERVER.show_message(message, lsp.MessageType.Error)


def log_warning(message: str) -> None:
    LSP_SERVER.show_message_log(message, lsp.MessageType.Warning)
    if os.getenv("LS_SHOW_NOTIFICATION", "off") in ["onWarning", "always"]:
        LSP_SERVER.show_message(message, lsp.MessageType.Warning)


def log_always(message: str) -> None:
    LSP_SERVER.show_message_log(message, lsp.MessageType.Info)
    if os.getenv("LS_SHOW_NOTIFICATION", "off") in ["always"]:
        LSP_SERVER.show_message(message, lsp.MessageType.Info)


# *****************************************************
# Start the server.
# *****************************************************
if __name__ == "__main__":
    LSP_SERVER.start_io()



"""
Common CLI writing utilities.
"""

import functools
import itertools
import logging
from pathlib import Path
from typing import Iterable

import click

import atopile.config
from atopile import address, errors, version
from atopile.address import AddrStr

log = logging.getLogger(__name__)


def project_options(f):
    """
    Utility decorator to ingest common config options to build a project.
    """

    @click.argument("entry", required=False, default=None)
    @click.option("-b", "--build", multiple=True, envvar="ATO_BUILD")
    @click.option("-t", "--target", multiple=True, envvar="ATO_TARGET")
    @click.option("-o", "--option", multiple=True, envvar="ATO_OPTION")
    @functools.wraps(f)
    @errors.muffle_fatalities
    def wrapper(
        *args,
        entry: str,
        build: Iterable[str],
        target: Iterable[str],
        option: Iterable[str],
        **kwargs,
    ):
        """Wrap a CLI command to ingest common config options to build a project."""

        # basic the entry address if provided, otherwise leave it as None
        if entry is not None:
            entry = AddrStr(entry)

            if address.get_file(entry) is None:
                raise click.BadParameter(
                    f"Invalid entry address {entry} - entry must specify a file.",
                    param_hint="entry",
                )

        # get the project
        if entry is None:
            entry_arg_file_path = Path.cwd()
        else:
            entry_arg_file_path = (
                Path(address.get_file(entry)).expanduser().resolve().absolute()
            )

        try:
            project_config = atopile.config.get_project_config_from_addr(str(entry_arg_file_path))
        except FileNotFoundError as ex:
            # FIXME: this raises an exception when the entry is not in a project
            raise click.BadParameter(
                f"Could not find project from path {str(entry_arg_file_path)}. "
                "Is this file path within a project?"
            ) from ex

        # Make sure I an all my sub-configs have appropriate versions
        check_compiler_versions(project_config)

        log.info("Using project %s", project_config.location)

        # add custom config overrides
        if option:
            raise NotImplementedError(
                "Custom config overrides have been removed in a refactor. "
                "It's planned to re-add them in a future release. "
                "If this is a blocker for you, please raise an issue. "
                "In the meantime, you can use the `ato.yaml` file to set these options."
            )
        else:
            config: atopile.config.ProjectConfig = project_config

        # if we set an entry-point, we now need to deal with that
        entry_addr_override = None
        if entry:
            if entry_arg_file_path.is_file():
                if entry_section := address.get_entry_section(entry):
                    entry_addr_override = address.from_parts(
                        str(entry_arg_file_path.absolute()),
                        entry_section,
                    )
                else:
                    raise click.BadParameter(
                        "If an entry of a file is specified, you must specify"
                        " the node within it you want to build.",
                        param_hint="entry",
                    )

            elif entry_arg_file_path.is_dir():
                pass

            elif not entry_arg_file_path.exists():
                raise click.BadParameter(
                    "The entry you have specified does not exist.",
                    param_hint="entry",
                )
            else:
                raise ValueError(
                    f"Unexpected entry path type {entry_arg_file_path} - this should never happen!"
                )

        # Configure project context
        project_ctx = atopile.config.ProjectContext.from_config(config)
        atopile.config.set_project_context(project_ctx)

        # Make build contexts
        with errors.handle_ato_errors():
            if build_names := build or config.builds.keys():
                build_ctxs: list[atopile.config.BuildContext] = [
                    atopile.config.BuildContext.from_config_name(config, build_name)
                    for build_name in build_names
                ]
            else:
                build_ctxs = [atopile.config.BuildContext.from_config(
                    "default",
                    atopile.config.ProjectBuildConfig(),
                    project_ctx
                )]

        for build_ctx in build_ctxs:
            if entry_addr_override is not None:
                build_ctx.entry = entry_addr_override
            if target:
                build_ctx.targets = list(target)

        # Do the thing
        return f(*args, **kwargs, build_ctxs=build_ctxs)

    return wrapper


def check_compiler_versions(config: atopile.config.ProjectConfig):
    """
    Check that the compiler version is compatible with the version
    used to build the project.
    """
    with errors.handle_ato_errors():
        dependency_cfgs = (
            errors.downgrade(atopile.config.get_project_config_from_path, FileNotFoundError)(p)
            for p in Path(config.location).glob("*")
        )

        for cltr, cfg in errors.iter_through_errors(
            itertools.chain([config], dependency_cfgs)
        ):
            if cfg is None:
                continue

            with cltr():
                semver_str = cfg.ato_version
                # FIXME: this is a hack to the moment to get around us breaking
                # the versioning scheme in the ato.yaml files
                for operator in version.OPERATORS:
                    semver_str = semver_str.replace(operator, "")

                built_with_version = version.parse(semver_str)

                if not version.match_compiler_compatability(built_with_version):
                    raise version.VersionMismatchError(
                        f"{cfg.location} ({cfg.ato_version}) can't be"
                        " built with this version of atopile "
                        f"({version.get_installed_atopile_version()})."
                    )



import json
import logging
from pathlib import Path
from typing import Any, Iterable

import pcbnew

LOG_FILE = Path("~/.atopile/kicad-plugin.log").expanduser().absolute()
LOG_FILE.parent.mkdir(parents=True, exist_ok=True)

log = logging.getLogger(__name__)
log.addHandler(logging.FileHandler(str(LOG_FILE), "w", "utf-8"))
log.setLevel(logging.DEBUG)


def get_prj_dir(path: Path) -> Path:
    """Return the atopile project directory."""
    path = Path(path)
    if (path / "ato.yaml").exists():
        return path
    for p in path.parents:
        if (p / "ato.yaml").exists():
            return p
    raise FileNotFoundError("ato.yaml not found in any parent directory")


def get_board_artifact_manifest(board_path: Path) -> dict:
    """Return a dict of the artifact manifest related to this board."""
    board_path = Path(board_path)
    manifest_path = get_prj_dir(board_path) / "build" / "manifest.json"
    with manifest_path.open("r") as f:
        manifest = json.load(f)
    return manifest.get("by-layout", {}).get(str(board_path), {})


def groups_by_name(board: pcbnew.BOARD) -> dict[str, pcbnew.PCB_GROUP]:
    """Return a dict of groups by name."""
    return {g.GetName(): g for g in board.Groups()}


def get_footprint_uuid(fp: pcbnew.FOOTPRINT) -> str:
    """Return the UUID of a footprint."""
    path = fp.GetPath().AsString()
    return path.split("/")[-1]


def footprints_by_uuid(board: pcbnew.BOARD) -> dict[str, pcbnew.FOOTPRINT]:
    """Return a dict of footprints by UUID."""
    return {get_footprint_uuid(fp): fp for fp in board.GetFootprints()}


def get_layout_map(board_path: Path) -> dict[str, Any]:
    """Return the layout map for the board."""
    board_path = Path(board_path)
    manifest = get_board_artifact_manifest(board_path)
    with Path(manifest["layouts"]).open("r", encoding="utf-8") as f:
        return json.load(f)


def flip_dict(d: dict) -> dict:
    """Return a dict with keys and values swapped."""
    return {v: k for k, v in d.items()}


def sync_track(
    source_board: pcbnew.BOARD,
    track: pcbnew.PCB_TRACK,
    target_board: pcbnew.BOARD
) -> pcbnew.PCB_TRACK:
    """Sync a track to the target board."""
    new_track: pcbnew.PCB_TRACK = track.Duplicate().Cast()
    new_track.SetParent(target_board)
    new_track.SetStart(track.GetStart())
    new_track.SetEnd(track.GetEnd())
    new_track.SetLayer(track.GetLayer())
    target_board.Add(new_track)
    source_board.Remove(new_track)
    return new_track


#TODO: There must be a better way to update net of fill
def update_zone_net(source_zone: pcbnew.ZONE, source_board: pcbnew.BOARD, target_zone: pcbnew.ZONE, target_board:pcbnew.BOARD, uuid_map: dict[str, str]):
    '''Finds a pin connected to original zone, pulls pin net name from new board net'''
    source_netname = source_zone.GetNetname()
    matched_fp = None
    matched_pad_index = None
    for fp in source_board.GetFootprints():
        for index, pad in enumerate(fp.Pads()):
            if pad.GetNetname() == source_netname:
                matched_fp = fp
                matched_pad_index = index

    if matched_fp and matched_pad_index:
        matched_fp_uuid = get_footprint_uuid(matched_fp)
        target_fp_uuid = uuid_map.get(matched_fp_uuid,None)
        if target_fp_uuid:
            target_uuids = footprints_by_uuid(target_board)
            target_fp = target_uuids.get(target_fp_uuid,None)
            if target_fp:
                new_netinfo = target_fp.Pads()[matched_pad_index].GetNet()
                target_zone.SetNet(new_netinfo)
                return

    # TODO: Verify that this will always set net to no net
    target_zone.SetNetCode(0)


def sync_zone(zone: pcbnew.ZONE, target: pcbnew.BOARD) -> pcbnew.ZONE:
    """Sync a zone to the target board."""
    new_zone: pcbnew.ZONE = zone.Duplicate().Cast()
    new_zone.SetParent(target)
    new_zone.SetLayer(zone.GetLayer())
    target.Add(new_zone)
    return new_zone


def sync_footprints(
    source: pcbnew.BOARD, target: pcbnew.BOARD, uuid_map: dict[str, str]
) -> list[str]:
    """Update the target board with the layout from the source board."""
    # Update the footprint position, orientation, and side to match the source
    source_uuids = footprints_by_uuid(source)
    target_uuids = footprints_by_uuid(target)
    missing_uuids = []
    for s_uuid, t_uuid in uuid_map.items():
        try:
            target_fp = target_uuids[t_uuid]
            source_fp = source_uuids[s_uuid]
        except KeyError:
            missing_uuids.append(s_uuid)
            continue
        target_fp.SetPosition(source_fp.GetPosition())
        target_fp.SetOrientation(source_fp.GetOrientation())
        target_fp.SetLayer(source_fp.GetLayer())
        # Ref Designators
        target_fp.Reference().SetAttributes(source_fp.Reference().GetAttributes())
        target_ref: pcbnew.FP_TEXT = target_fp.Reference()
        source_ref: pcbnew.FP_TEXT = source_fp.Reference()
        target_ref.SetPosition(source_ref.GetPosition())

    return missing_uuids


def find_anchor_footprint(fps: Iterable[pcbnew.FOOTPRINT]) -> pcbnew.FOOTPRINT:
    """Return anchor footprint with largest pin count in board or group: tiebreaker size"""
    # fps = layout.GetFootprints()
    # fps = [item for item in layout.GetItems() if isinstance(item, pcbnew.FOOTPRINT)]
    max_padcount = 0
    max_area = 0
    anchor_fp = None
    for fp in fps:
        if fp.GetPadCount() > max_padcount or (fp.GetPadCount()==max_padcount and fp.GetArea()>max_area):
            anchor_fp = fp
            max_padcount = fp.GetPadCount()
            max_area = fp.GetArea()
    return anchor_fp


def get_group_footprints(group: pcbnew.PCB_GROUP) -> list[pcbnew.FOOTPRINT]:
    """Return a list of footprints in a group."""
    return [item for item in group.GetItems() if isinstance(item, pcbnew.FOOTPRINT)]


def calculate_translation(source_fps: Iterable[pcbnew.FOOTPRINT], target_fps: Iterable[pcbnew.FOOTPRINT]) -> pcbnew.VECTOR2I:
    """Calculate the translation vector between two groups of footprints."""
    source_anchor_fp = find_anchor_footprint(source_fps)
    source_offset = source_anchor_fp.GetPosition()
    target_anchor_fp = find_anchor_footprint(target_fps)
    target_offset = target_anchor_fp.GetPosition()
    total_offset = target_offset - source_offset
    return total_offset



from atopile import config
from ruamel.yaml import YAML
import copy

yaml = YAML()


def test_roundtrip():
    config_dict = yaml.load("""
        ato-version: ^0.2.0
        builds:
            debug:
                entry: elec/src/debug.ato:Debug
        unknown: test
        # comments
        dependencies:
        - tps63020dsjr # comments
        - usb-connectors ^v2.0.1
        - name: esp32-s3
          version: ^v0.0.1
          path: ../esp32-s3
        """)
    cfg = config.ProjectConfig.structure(config_dict)
    assert config_dict == cfg.patch_config(config_dict)
    assert cfg.ato_version == "^0.2.0"

    # Make some changes and check that they are reflected in the config
    cfg.ato_version = "10"
    cfg.dependencies[1].path = "test"
    config_dict_2 = copy.deepcopy(config_dict)
    config_dict_2["ato-version"] = "10"
    config_dict_2["dependencies"][2] = {'name': 'esp32-s3', 'version_spec': None, 'link_broken': False, 'path': '../esp32-s3'}
    config_dict_2["dependencies"][1] = {'name': 'usb-connectors', 'version': '^v0.0.1', 'path': 'test'}
    assert config_dict_2 == cfg.patch_config(config_dict)



# pylint: disable=too-few-public-methods

"""Schema and utils for atopile config files."""

import copy
import fnmatch
import logging
from pathlib import Path
from typing import Optional

import cattrs
import deepdiff
from attrs import Factory, define
from ruamel.yaml import YAML

import atopile.errors
import atopile.version
from atopile import address

log = logging.getLogger(__name__)
yaml = YAML()


CONFIG_FILENAME = "ato.yaml"
ATO_DIR_NAME = ".ato"
MODULE_DIR_NAME = "modules"
BUILD_DIR_NAME = "build"


_converter = cattrs.Converter()


class AtoConfigError(atopile.errors.AtoError):
    """An error in the config file."""


@define
class ProjectPaths:
    """Config grouping for all the paths in a project."""

    src: Path = "elec/src"
    layout: Path = "elec/layout"


@define
class ProjectBuildConfig:
    """Config for a build."""

    entry: Optional[str] = None
    targets: list[str] = ["__default__"]


@define
class ProjectServicesConfig:
    """A config for services used by the project."""
    components: str = "https://component-server-3033-5335559d-kjaci698.onporter.run/jlc/v1"


@define
class Dependency:
    """A dependency for a project."""

    name: str
    version_spec: Optional[str] = None
    link_broken: bool = False
    path: Optional[Path] = None

    @classmethod
    def from_str(cls, spec_str: str) -> "Dependency":
        """Create a Dependency object from a string."""
        for splitter in atopile.version.OPERATORS + ("@",):
            if splitter in spec_str:
                try:
                    name, version_spec = spec_str.split(splitter)
                    name = name.strip()
                    version_spec = version_spec.strip()
                    version_spec = splitter + version_spec
                except TypeError as ex:
                    raise atopile.errors.AtoTypeError(
                        f"Invalid dependency spec: {spec_str}"
                    ) from ex
                return cls(name, version_spec)
        return cls(name=spec_str)


@define
class ProjectConfig:
    """
    The config object for atopile.
    """

    location: Optional[Path] = None

    ato_version: str = "0.1.0"
    paths: ProjectPaths = Factory(ProjectPaths)
    builds: dict[str, ProjectBuildConfig] = Factory(dict)
    dependencies: list[str | Dependency] = Factory(list)
    services: ProjectServicesConfig = Factory(ProjectServicesConfig)

    @staticmethod
    def _sanitise_dict_keys(d: dict) -> dict:
        """Sanitise the keys of a dictionary to be valid python identifiers."""
        data = copy.deepcopy(d)
        data["ato_version"] = data.pop("ato-version")
        return data

    @staticmethod
    def _unsanitise_dict_keys(d: dict) -> dict:
        """Sanitise the keys of a dictionary to be valid python identifiers."""
        data = copy.deepcopy(d)
        data["ato-version"] = data.pop("ato_version")
        del data["location"]  # The location is saved by the literal location of the file
        return data

    @classmethod
    def structure(cls, data: dict) -> "ProjectConfig":
        """Make a config object from a dictionary."""
        try:
            return _converter.structure(cls._sanitise_dict_keys(data), cls)
        except* KeyError as exs:
            for ex in exs.exceptions:
                # FIXME: make this less shit
                raise AtoConfigError(f"Bad key in config {repr(ex)}") from ex

    def patch_config(self, original: dict) -> dict:
        """Apply a delta between the original and the current config."""
        original_cfg = self.structure(original)

        # Here we need to work around some structural changes
        # FIXME: the ideal behaviour here would be to default back to
        # the new structure whenever there's a conflict, but I can't
        # find a hook to callback to a "conflict-resolver" or the likes
        # and the exceptions don't have sufficient information to easily find them
        original_deps_by_name = {d.name: d for d in original_cfg.dependencies}
        original_dep_indicies = {d.name: i for i, d in enumerate(original_cfg.dependencies)}
        for d in self.dependencies:
            if d.name in original_deps_by_name and d != original_deps_by_name[d.name]:
                del original["dependencies"][original_dep_indicies[d.name]]
        original_cfg = self.structure(original)
        # Kill me... I'm sorry

        diff = deepdiff.DeepDiff(
            self._unsanitise_dict_keys(_converter.unstructure(original_cfg)),
            self._unsanitise_dict_keys(_converter.unstructure(self)),
        )

        delta = deepdiff.Delta(diff)
        return original + delta

    def save_changes(
        self,
        location: Optional[Path] = None
    ) -> None:
        """
        Save the changes to the config object
        """
        if location is None:
            location = self.location / CONFIG_FILENAME

        with location.open() as f:
            original = yaml.load(f)

        patched = self.patch_config(original)

        with location.open("w") as f:
            yaml.dump(patched, f)

    @classmethod
    def load(cls, location: Path) -> "ProjectConfig":
        """
        Make a config object for a project.
        """
        with location.open() as f:
            config_data = yaml.load(f)

        config = cls.structure(config_data)
        config.location = location.parent.expanduser().resolve().absolute()

        return config


## Register hooks for cattrs to handle the custom types

_converter.register_structure_hook(
    str | Dependency,
    lambda d, _: Dependency.from_str(d) if isinstance(d, str) else _converter.structure(d, Dependency),
)

##


def get_project_dir_from_path(path: Path) -> Path:
    """
    Resolve the project directory from the specified path.
    """
    #TODO: when provided with the "." path, it doesn't find the config in parent directories
    path = Path(path)
    for p in [path] + list(path.parents):
        clean_path = p.resolve().absolute()
        if (clean_path / CONFIG_FILENAME).exists():
            return clean_path
    raise atopile.errors.AtoFileNotFoundError(
        f"Could not find {CONFIG_FILENAME} in {path} or any parents"
    )


_loaded_configs: dict[Path, str] = {}


def get_project_config_from_path(path: Path) -> ProjectConfig:
    """
    Get the project config from a path.
    """
    project_dir = get_project_dir_from_path(path)
    project_config_file = project_dir / CONFIG_FILENAME
    if project_config_file not in _loaded_configs:
        _loaded_configs[project_config_file] = ProjectConfig.load(project_config_file)
    return _loaded_configs[project_config_file]


def get_project_config_from_addr(addr: str) -> ProjectConfig:
    """
    Get the project config from an address.
    """
    return get_project_config_from_path(Path(address.get_file(addr)))


# FIXME: we need factory constructors for these classes
@define
class ProjectContext:
    """A class to hold the arguments to a project."""

    project_path: Path  # abs path to the project directory
    src_path: Path  # abs path to the source directory
    module_path: Path  # abs path to the module directory
    layout_path: Path  # eg. path/to/project/layouts/default/default.kicad_pcb
    config: ProjectConfig

    @classmethod
    def from_config(cls, config: ProjectConfig) -> "ProjectContext":
        """Create a BuildArgs object from a Config object."""

        return ProjectContext(
            project_path=Path(config.location),
            src_path=Path(config.location) / config.paths.src,
            module_path=Path(config.location) / ATO_DIR_NAME / MODULE_DIR_NAME,
            layout_path=Path(config.location) / config.paths.layout,
            config=config,
        )

    @classmethod
    def from_path(cls, path: Path) -> "ProjectContext":
        """Create a BuildArgs object from a Config object."""
        return cls.from_config(get_project_config_from_path(path))


def match_user_layout(path: Path) -> bool:
    """Check whether a given filename is a KiCAD autosaved layout."""
    autosave_patterns = [
        "_autosave-*",
        "*-save.kicad_pcb",
    ]

    name = path.name

    for pattern in autosave_patterns:
        if fnmatch.fnmatch(name, pattern):
            return False
    return True


def find_layout(layout_base: Path) -> Optional[Path]:
    """Return the layout associated with a build."""
    if layout_base.with_suffix(".kicad_pcb").exists():
        return layout_base.with_suffix(".kicad_pcb").resolve().absolute()
    elif layout_base.is_dir():
        layout_candidates = list(
            filter(
                match_user_layout,
                layout_base.glob("*.kicad_pcb")
            )
        )

        if len(layout_candidates) == 1:
            return layout_candidates[0].resolve().absolute()

        else:
            raise atopile.errors.AtoError(
                "Layout directories must contain only 1 layout,"
                f" but {len(layout_candidates)} found in {layout_base}"
            )


@define
class BuildContext:
    """A class to hold the arguments to a build."""
    project_context: ProjectContext

    name: str

    entry: address.AddrStr  # eg. "path/to/project/src/entry-name.ato:module.path"
    targets: list[str]

    layout_path: Optional[Path]  # eg. path/to/project/layouts/default/default.kicad_pcb
    build_path: Path  # eg. path/to/project/build/<build-name>

    output_base: Path  # eg. path/to/project/build/<build-name>/entry-name

    @classmethod
    def from_config(
        cls,
        config_name: str,
        build_config: ProjectBuildConfig,
        project_context: ProjectContext
    ) -> "BuildContext":
        """Create a BuildArgs object from a Config object."""
        abs_entry = address.AddrStr(project_context.project_path / build_config.entry)

        build_path = Path(project_context.project_path) / BUILD_DIR_NAME

        return BuildContext(
            project_context=project_context,
            name=config_name,
            entry=abs_entry,
            targets=build_config.targets,
            layout_path=find_layout(project_context.project_path / project_context.layout_path / config_name),
            build_path=build_path,
            output_base=build_path / config_name,
        )

    @classmethod
    def from_config_name(cls, config: ProjectConfig, build_name: str) -> "BuildContext":
        """Create a BuildArgs object from a Config object."""
        project_context = ProjectContext.from_config(config)

        try:
            build_config = config.builds[build_name]
        except KeyError as ex:
            raise atopile.errors.AtoError(
                f"Build {build_name} not found for project {config.location}\n"
                f"Available builds: {list(config.builds.keys())}"
            ) from ex

        return cls.from_config(build_name, build_config, project_context)


_project_context: Optional[ProjectContext] = None


def set_project_context(project_context: ProjectContext) -> None:
    """
    Set the project context for the current process.
    """
    global _project_context
    _project_context = project_context


def get_project_context() -> ProjectContext:
    """
    Get the project context for the current process.
    """
    if _project_context is None:
        raise ValueError("Project context not set")
    return _project_context
